# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A CLI tool for dfindexeddb."""
import argparse
import dataclasses
from datetime import datetime
import json
import pathlib
import sys
import traceback

from dfindexeddb import errors
from dfindexeddb import version
from dfindexeddb.leveldb import record as leveldb_record
from dfindexeddb.indexeddb import chromium
from dfindexeddb.indexeddb import v8


_VALID_PRINTABLE_CHARACTERS = (
    ' abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789' +
    '!"#$%&\'()*+,-./:;<=>?@[\\]^_`{|}~.')


class Encoder(json.JSONEncoder):
  """A JSON encoder class for dfindexeddb fields."""
  def default(self, o):
    if dataclasses.is_dataclass(o):
      o_dict = dataclasses.asdict(o)
      return o_dict
    if isinstance(o, bytes):
      out = []
      for x in o:
        if chr(x) not in _VALID_PRINTABLE_CHARACTERS:
          out.append(f'\\x{x:02X}')
        else:
          out.append(chr(x))
      return ''.join(out)
    if isinstance(o, datetime):
      return o.isoformat()
    if isinstance(o, v8.Undefined):
      return "<undefined>"
    if isinstance(o, v8.Null):
      return "<null>"
    if isinstance(o, set):
      return list(o)
    if isinstance(o, v8.RegExp):
      return str(o)
    return json.JSONEncoder.default(self, o)


def _Output(structure, to_json=False):
  """Helper method to output parsed structure to stdout."""
  if to_json:
    print(json.dumps(structure, indent=2, cls=Encoder))
  else:
    print(structure)


def IndexeddbCommand(args):
  """The CLI for processing a log/ldb file as indexeddb."""
  for db_record in leveldb_record.LevelDBRecord.FromDir(args.source):
    record = db_record.record
    try:
      db_record.record = chromium.IndexedDBRecord.FromLevelDBRecord(record)
    except(
        errors.ParserError,
        errors.DecoderError,
        NotImplementedError) as err:
      print(
          (f'Error parsing blink value: {err} for {record.__class__.__name__} '
           f'at offset {record.offset} in {db_record.path}'), file=sys.stderr)
      print(f'Traceback: {traceback.format_exc()}', file=sys.stderr)
    _Output(db_record, to_json=args.json)


def App():
  """The CLI app entrypoint for dfindexeddb."""
  parser = argparse.ArgumentParser(
      prog='dfindexeddb',
      description='A cli tool for parsing indexeddb files',
      epilog=f'Version {version.GetVersion()}')
  parser.add_argument(
      '-s', '--source', required=True, type=pathlib.Path,
      help='The source leveldb folder')
  parser.add_argument('--json', action='store_true', help='Output as JSON')
  parser.set_defaults(func=IndexeddbCommand)

  args = parser.parse_args()
  args.func(args)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parser for LevelDB Descriptor (MANIFEST) files."""
from __future__ import annotations

from dataclasses import dataclass, field
from typing import Generator, Optional

from dfindexeddb import errors
from dfindexeddb.leveldb import definitions
from dfindexeddb.leveldb import log
from dfindexeddb.leveldb import utils


@dataclass
class InternalKey:
  """An InternalKey.

  Attributes:
    offset: the offset.
    user_key: the user key.
    sequence_number: the sequence number.
    key_type: the key type.
  """
  offset: int
  user_key: bytes = field(repr=False)
  sequence_number: int
  key_type: int

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0) -> InternalKey:
    """Decodes an InternalKey from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The InternalKey instance.
    """
    offset, slice_bytes = decoder.DecodeLengthPrefixedSlice()

    if len(slice_bytes) < definitions.PACKED_SEQUENCE_AND_TYPE_LENGTH:
      raise errors.ParserError('Insufficient bytes to parse InternalKey')

    user_key = slice_bytes[:-definitions.SEQUENCE_LENGTH]
    sequence_number = int.from_bytes(
        slice_bytes[-definitions.SEQUENCE_LENGTH:],
        byteorder='little',
        signed=False)
    key_type = slice_bytes[-definitions.PACKED_SEQUENCE_AND_TYPE_LENGTH]

    return cls(
        offset=base_offset + offset,
        user_key=user_key,
        sequence_number=sequence_number,
        key_type=key_type)


@dataclass
class NewFile(utils.FromDecoderMixin):
  """A NewFile.

  Attributes:
    offset: the offset.
    level: the level.
    number: the file number.
    file_size: the file size.
    smallest: the smallest internal key.
    largest: the largest internal key.
  """
  offset: int
  level: int
  number: int
  file_size: int
  smallest: InternalKey
  largest: InternalKey

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0) -> NewFile:
    """Decodes a NewFile from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The NewFile instance.
    """
    offset, level = decoder.DecodeUint32Varint()
    _, number = decoder.DecodeUint64Varint()
    _, file_size = decoder.DecodeUint64Varint()
    smallest = InternalKey.FromDecoder(decoder, base_offset=base_offset)
    largest = InternalKey.FromDecoder(decoder, base_offset=base_offset)

    return cls(
        offset=offset + base_offset,
        level=level,
        number=number,
        file_size=file_size,
        smallest=smallest,
        largest=largest)


@dataclass
class CompactPointer(utils.FromDecoderMixin):
  """A CompactPointer.

  Attributes:
    offset: the offset.
    level: the level.
    key: the key bytes.
  """
  offset: int
  level: int
  key: bytes = field(repr=False)

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0
  ) -> CompactPointer:
    """Decodes a CompactPointer from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The CompactPointer instance.
    """
    offset, level = decoder.DecodeUint32Varint()
    _, key = decoder.DecodeLengthPrefixedSlice()
    return cls(offset=base_offset + offset, level=level, key=key)


@dataclass
class DeletedFile(utils.FromDecoderMixin):
  """A DeletedFile.

  Attributes:
    offset: the offset.
    level: the level.
    number: the file number.
  """
  offset: int
  level: int
  number: int

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0) -> DeletedFile:
    """Decodes a DeletedFile from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The DeletedFile instance.
    """
    offset, level = decoder.DecodeUint32Varint()
    _, number = decoder.DecodeUint64Varint()
    return cls(offset=base_offset + offset, level=level, number=number)


@dataclass
class VersionEdit(utils.FromDecoderMixin):
  """A VersionEdit is recorded in a LevelDB descriptor/manifest file.

  Attributes:
    offset: the offset where the VersionEdit was parsed.
    comparator: the comparator.
    log_number: the log number.
    prev_log_number: the previous log number.
    next_file_number: the next file number.
    last_sequence: the last sequence.
    compact_pointers: the list of CompactPointers.
    deleted_files: the list of DeletedFiles.
    new_files: the list of NewFiles.
  """
  offset: int
  comparator: Optional[bytes] = None
  log_number: Optional[int] = None
  prev_log_number: Optional[int] = None
  next_file_number: Optional[int] = None
  last_sequence: Optional[int] = None
  compact_pointers: list[CompactPointer] = field(default_factory=list)
  deleted_files: list[DeletedFile] = field(default_factory=list)
  new_files: list[NewFile] = field(default_factory=list)

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0) -> VersionEdit:
    """Decodes a VersionEdit from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The VersionEdit instance.

    Raises:
      ParserError if an invalid VersionEditTag is parsed.
    """
    offset, tag_byte = decoder.DecodeUint32Varint()
    version_edit = cls(offset=base_offset + offset)

    while tag_byte:
      try:
        tag = definitions.VersionEditTags(tag_byte)
      except TypeError as error:
        raise errors.ParserError(
            f'Invalid VersionEditTag at offset {offset}') from error

      if tag == definitions.VersionEditTags.COMPARATOR:
        _, version_edit.comparator = decoder.DecodeLengthPrefixedSlice()
      elif tag == definitions.VersionEditTags.LOG_NUMBER:
        _, version_edit.log_number = decoder.DecodeUint64Varint()
      elif tag == definitions.VersionEditTags.PREV_LOG_NUMBER:
        _, version_edit.prev_log_number = decoder.DecodeUint64Varint()
      elif tag == definitions.VersionEditTags.NEXT_FILE_NUMBER:
        _, version_edit.next_file_number = decoder.DecodeUint64Varint()
      elif tag == definitions.VersionEditTags.LAST_SEQUENCE:
        _, version_edit.last_sequence = decoder.DecodeUint64Varint()
      elif tag == definitions.VersionEditTags.COMPACT_POINTER:
        compact_pointer = CompactPointer.FromDecoder(
            decoder=decoder, base_offset=base_offset + offset)
        version_edit.compact_pointers.append(compact_pointer)
      elif tag == definitions.VersionEditTags.DELETED_FILE:
        deleted_file = DeletedFile.FromDecoder(
            decoder=decoder, base_offset=base_offset + offset)
        version_edit.deleted_files.append(deleted_file)
      elif tag == definitions.VersionEditTags.NEW_FILE:
        file_metadata = NewFile.FromDecoder(
            decoder=decoder, base_offset=base_offset + offset)
        version_edit.new_files.append(file_metadata)

      if decoder.NumRemainingBytes() == 0:
        break

      offset, tag_byte = decoder.DecodeUint32Varint()

    return version_edit


class FileReader:
  """A reader for Descriptor files.

  A DescriptorFileReader provides read-only sequential iteration of serialized
  structures in a leveldb Descriptor file.  These structures include:
  * blocks (Block)
  * records (PhysicalRecord)
  * version edits (VersionEdit)
  """
  def __init__(self, filename: str):
    """Initializes the Descriptor a.k.a. MANIFEST file.

    Args:
      filename: the Descriptor filename (e.g. MANIFEST-000001)
    """
    self.filename = filename


  def GetBlocks(self) -> Generator[log.Block, None, None]:
    """Returns an iterator of Block instances.

    A Descriptor file is composed of one or more blocks.

    Yields:
      Block
    """
    with open(self.filename, 'rb') as fh:
      block = log.Block.FromStream(fh)
      while block:
        yield block
        block = log.Block.FromStream(fh)

  def GetPhysicalRecords(self) -> Generator[log.PhysicalRecord, None, None]:
    """Returns an iterator of PhysicalRecord instances.

    A block is composed of one or more physical records.

    Yields:
      PhysicalRecord
    """
    for block in self.GetBlocks():
      yield from block.GetPhysicalRecords()

  def GetVersionEdits(self) -> Generator[VersionEdit, None, None]:
    """Returns an iterator of VersionEdit instances.

    Depending on the VersionEdit size, it can be spread across one or
    more physical records.

    Yields:
      VersionEdit
    """
    buffer = bytearray()
    for physical_record in self.GetPhysicalRecords():
      if (physical_record.record_type ==
          definitions.LogFilePhysicalRecordType.FULL):
        buffer = physical_record.contents
        offset = physical_record.contents_offset + physical_record.base_offset
        version_edit = VersionEdit.FromBytes(buffer, base_offset=offset)
        yield version_edit
        buffer = bytearray()
      elif (physical_record.record_type ==
            definitions.LogFilePhysicalRecordType.FIRST):
        offset = physical_record.contents_offset + physical_record.base_offset
        buffer = bytearray(physical_record.contents)
      elif (physical_record.record_type ==
            definitions.LogFilePhysicalRecordType.MIDDLE):
        buffer.extend(bytearray(physical_record.contents))
      elif (physical_record.record_type ==
            definitions.LogFilePhysicalRecordType.LAST):
        buffer.extend(bytearray(physical_record.contents))
        version_edit = VersionEdit.FromBytes(buffer, base_offset=offset)
        yield version_edit
        buffer = bytearray()



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Definitions for LevelDB."""

import enum

BLOCK_RESTART_ENTRY_LENGTH = 4
BLOCK_TRAILER_SIZE = 5
TABLE_FOOTER_SIZE = 48
TABLE_MAGIC = b'\x57\xfb\x80\x8b\x24\x75\x47\xdb'

PACKED_SEQUENCE_AND_TYPE_LENGTH = 8
SEQUENCE_LENGTH = 7
TYPE_LENGTH = 1


class BlockCompressionType(enum.IntEnum):
  """Block compression types."""
  SNAPPY = 1
  ZSTD = 2


class VersionEditTags(enum.IntEnum):
  """VersionEdit tags."""
  COMPARATOR = 1
  LOG_NUMBER = 2
  NEXT_FILE_NUMBER = 3
  LAST_SEQUENCE = 4
  COMPACT_POINTER = 5
  DELETED_FILE = 6
  NEW_FILE = 7
  # 8 was used for large value refs
  PREV_LOG_NUMBER = 9


class LogFilePhysicalRecordType(enum.IntEnum):
  """Log file physical record types."""
  FULL = 1
  FIRST = 2
  MIDDLE = 3
  LAST = 4


class InternalRecordType(enum.IntEnum):
  """Internal record types."""
  DELETED = 0
  VALUE = 1



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Helper/utility classes for LevelDB."""
from __future__ import annotations
import io
from typing import BinaryIO, Tuple, Type, TypeVar

from dfindexeddb import errors
from dfindexeddb import utils


class LevelDBDecoder(utils.StreamDecoder):
  """A helper class to decode data types from LevelDB files."""

  def DecodeBool(self) -> Tuple[int, bool]:
    """Returns a Tuple of the offset of decoding and the bool value."""
    offset, buffer = self.ReadBytes(1)
    return offset, buffer[0] is not None

  def DecodeString(self) -> Tuple[int, str]:
    """Returns a tuple of the offset of decoding and the string value.

    Raises:
      errors.DecoderError: when the parsed string buffer is not even (i.e.
          cannot be decoded as a UTF-16-BE string.
    """
    offset = self.stream.tell()
    buffer = self.stream.read()
    if len(buffer) % 2:
      raise errors.DecoderError(
          f'Odd number of bytes encountered at offset {offset}')
    return offset, buffer.decode('utf-16-be')

  def DecodeLengthPrefixedSlice(self) -> Tuple[int, bytes]:
    """Returns a tuple of the offset of decoding and the byte 'slice'."""
    offset, num_bytes = self.DecodeUint32Varint()
    _, blob = self.ReadBytes(num_bytes)
    return offset, blob

  def DecodeBlobWithLength(self) -> Tuple[int, bytes]:
    """Returns a tuple of a the offset of decoding and the binary blob."""
    offset, num_bytes = self.DecodeUint64Varint()
    _, blob = self.ReadBytes(num_bytes)
    return offset, blob

  def DecodeStringWithLength(self, encoding='utf-16-be') -> Tuple[int, str]:
    """Returns a tuple of the offset of decoding and the string value."""
    offset, length = self.DecodeUint64Varint()
    _, buffer = self.ReadBytes(length*2)
    return offset, buffer.decode(encoding=encoding)


T = TypeVar('T')


class FromDecoderMixin:
  """A mixin for parsing dataclass attributes using a LevelDBDecoder."""

  @classmethod
  def FromDecoder(
      cls: Type[T], decoder: LevelDBDecoder, base_offset: int = 0) -> T:
    """Decodes a class type from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset.

    Returns:
      The class instance.

    Raises:
      NotImplementedError if the child class does not implement this method.
    """
    raise NotImplementedError

  @classmethod
  def FromStream(
      cls: Type[T], stream: BinaryIO, base_offset: int = 0) -> T:
    """Decodes a class type from the current position of a binary stream.

    Args:
      stream: the binary stream.
      base_offset: the base offset of the binary stream.

    Returns:
      The class instance.
    """
    decoder = LevelDBDecoder(stream)
    return cls.FromDecoder(decoder=decoder, base_offset=base_offset)

  @classmethod
  def FromBytes(
      cls: Type[T], raw_data: bytes, base_offset: int = 0) -> T:
    """Parses a class type from raw bytes.

    Args:
      raw_data: the raw data.
      base_offset: the base offset of the raw data.

    Returns:
      The class instance.
    """
    stream = io.BytesIO(raw_data)
    return cls.FromStream(stream=stream, base_offset=base_offset)



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A module for records from LevelDB files."""
from __future__ import annotations
import dataclasses
import pathlib
import sys
from typing import Any, Generator, Union

from dfindexeddb.leveldb import descriptor
from dfindexeddb.leveldb import ldb
from dfindexeddb.leveldb import log


@dataclasses.dataclass
class LevelDBRecord:
  """A leveldb record.

  A record can come from a log file, a table (ldb) file or a descriptor
  (MANIFEST) file.

  Attributes:
    path: the file path where the record was parsed from.
    record: the leveldb record.
  """
  path: str
  record: Union[
      ldb.KeyValueRecord,
      log.ParsedInternalKey,
      descriptor.VersionEdit]

  @classmethod
  def FromFile(
      cls,
      file_path: pathlib.Path,
      include_versionedit: bool = False
  ) -> Generator[LevelDBRecord, Any, Any]:
    """Yields leveldb records from the given path.

    Yields:
      LevelDBRecords

    Args:
      file_path: the file path.
      include_versionedit: include VersionEdit records from descriptor files.
    """
    if file_path.name.endswith('.log'):
      for record in log.FileReader(
          file_path.as_posix()).GetParsedInternalKeys():
        yield cls(path=file_path.as_posix(), record=record)
    elif file_path.name.endswith('.ldb'):
      for record in ldb.FileReader(file_path.as_posix()).GetKeyValueRecords():
        yield cls(path=file_path.as_posix(), record=record)
    elif file_path.name.startswith('MANIFEST'):
      if not include_versionedit:
        print(f'Ignoring {file_path.as_posix()}', file=sys.stderr)
        return
      for record in descriptor.FileReader(
          file_path.as_posix()).GetVersionEdits():
        yield cls(path=file_path.as_posix(), record=record)
    elif file_path.name in ('LOCK', 'CURRENT', 'LOG', 'LOG.old'):
      print(f'Ignoring {file_path.as_posix()}', file=sys.stderr)
    else:
      print(f'Unsupported file type {file_path.as_posix()}', file=sys.stderr)

  @classmethod
  def FromDir(
      cls,
      path: pathlib.Path,
      include_versionedit: bool = False
  ) -> Generator[LevelDBRecord, Any, Any]:
    """Yields LevelDBRecords from the given directory.

    Args:
      path: the file path.
      include_versionedit: include VersionEdit records from descriptor files.

    Yields:
      LevelDBRecords

    Raises:
      ValueError: if path is not a directory.
    """
    if path.is_dir():
      for file_path in path.iterdir():
        yield from cls.FromFile(
            file_path=file_path,
            include_versionedit=include_versionedit)
    else:
      raise ValueError(f'{path} is not a directory')



# -*- coding: utf-8 -*-
# Copyright 2024 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Parser for LevelDB Log (.log) files."""
from __future__ import annotations

from dataclasses import dataclass
import io
from typing import BinaryIO, Generator, Iterable, Optional

from dfindexeddb import errors
from dfindexeddb.leveldb import definitions
from dfindexeddb.leveldb import utils


@dataclass
class ParsedInternalKey:
  """An internal key record from a leveldb log file.

  Attributes:
    offset: the offset of the record.
    record_type: the record type.
    sequence_number: the sequence number (inferred from the relative location
        the ParsedInternalKey in a WriteBatch.)
    key: the record key.
    value: the record value.
  """
  offset: int
  record_type: definitions.InternalRecordType
  sequence_number: int
  key: bytes
  value: bytes

  @classmethod
  def FromDecoder(
      cls,
      decoder: utils.LevelDBDecoder,
      base_offset: int = 0,
      sequence_number: int = 0,
  ) -> ParsedInternalKey:
    """Decodes an internal key value record.

    Args:
      decoder: the leveldb decoder.
      base_offset: the base offset for the parsed internal key value record.
      sequence_number: the sequence number for the parsed internal key value
          record.

    Returns:
      A ParsedInternalKey

    Raises:
      ValueError: if there is an invalid record type encountered.
    """
    offset, record_type = decoder.DecodeUint8()
    _, key = decoder.DecodeBlobWithLength()
    record_type = definitions.InternalRecordType(record_type)

    if record_type == definitions.InternalRecordType.VALUE:
      _, value = decoder.DecodeBlobWithLength()
    elif record_type == definitions.InternalRecordType.DELETED:
      value =  b''
    else:
      raise ValueError(f'Invalid record type {record_type}')
    return cls(
        offset=base_offset + offset,
        record_type=record_type,
        key=key,
        value=value,
        sequence_number=sequence_number)


@dataclass
class WriteBatch(utils.FromDecoderMixin):
  """A write batch from a leveldb log file.

  Attributes:
    offset: the write batch offset.
    sequence_number: the batch sequence number.
    count: the number of ParsedInternalKey in the batch.
    records: the ParsedInternalKey parsed from the batch.
  """
  offset: int
  sequence_number: int
  count: int
  records: Iterable[ParsedInternalKey]

  @classmethod
  def FromDecoder(
    cls, decoder: utils.LevelDBDecoder, base_offset: int = 0
  ) -> WriteBatch:
    """Parses a WriteBatch from a binary stream.

    Args:
      decoder: the LevelDBDecoder
      base_offset: the base offset of the Block from which the data is
          read from.

    Returns:
      A WriteBatch.
    """
    offset, sequence_number = decoder.DecodeUint64()
    _, count = decoder.DecodeUint32()

    records = []
    for relative_sequence_number in range(count):
      record = ParsedInternalKey.FromDecoder(
          decoder, base_offset + offset,
          relative_sequence_number + sequence_number
      )
      records.append(record)
    return cls(
        offset=base_offset + offset,
        sequence_number=sequence_number,
        count=count,
        records=records)


@dataclass
class PhysicalRecord(utils.FromDecoderMixin):
  """A physical record from a leveldb log file.

  Attributes:
    offset: the record offset.
    checksum: the record checksum.
    length: the length of the record in bytes.
    type: the record type.
    contents: the record contents.
    contents_offset: the offset of where the record contents are stored.
  """
  base_offset: int
  offset: int
  checksum: int
  length: int
  record_type: definitions.LogFilePhysicalRecordType
  contents: bytes
  contents_offset: int

  PHYSICAL_HEADER_LENGTH = 7

  @classmethod
  def FromDecoder(
      cls, decoder: utils.LevelDBDecoder, base_offset: int = 0
  ) -> PhysicalRecord:
    """Decodes a PhysicalRecord from the current position of a LevelDBDecoder.

    Args:
      decoder: the LevelDBDecoder.
      base_offset: the base offset of the WriteBatch from which the data is
          read from.

    Returns:
      A PhysicalRecord.
    """
    offset, checksum = decoder.DecodeUint32()
    _, length = decoder.DecodeUint16()
    _, record_type_byte = decoder.DecodeUint8()
    try:
      record_type = definitions.LogFilePhysicalRecordType(record_type_byte)
    except ValueError as error:
      raise errors.ParserError(
          f'Error parsing record type of Physical Record at offset '
          f'{offset + base_offset}') from error
    contents_offset, contents = decoder.ReadBytes(length)
    return cls(
        base_offset=base_offset,
        offset=offset,
        checksum=checksum,
        length=length,
        record_type=record_type,
        contents=contents,
        contents_offset=contents_offset)


@dataclass
class Block:
  """A block from a leveldb log file.

  Attributes:
    offset: the block offset.
    data: the block data.
  """
  offset: int
  data: bytes

  BLOCK_SIZE = 32768

  def GetPhysicalRecords(self) -> Generator[PhysicalRecord, None, None]:
    """Returns a generator of LogFilePhysicalRecords from the block contents.

    Yields:
      LogFileRecord
    """
    buffer = io.BytesIO(self.data)
    buffer_length = len(self.data)

    while buffer.tell() + PhysicalRecord.PHYSICAL_HEADER_LENGTH < buffer_length:
      yield PhysicalRecord.FromStream(buffer, base_offset=self.offset)

  @classmethod
  def FromStream(cls, stream: BinaryIO) -> Optional[Block]:
    """Parses a Block from a binary stream.

    Args:
      stream: the binary stream to be parsed.

    Returns:
      the Block or None if there is no data to read from the stream.
    """
    offset = stream.tell()
    data = stream.read(cls.BLOCK_SIZE)  # reads full and partial blocks
    if not data:
      return None
    return cls(offset, data)


class FileReader:
  """A leveldb log file reader.

  A Log FileReader provides read-only sequential iteration of serialized
  structures in a leveldb logfile.  These structures include:
  * blocks (Block)
  * phyiscal records (PhysicalRecord)
  * batches (WriteBatch) and
  * key/value records (ParsedInternalKey).

  Attributes:
    filename (str): the leveldb log filename.
  """

  def __init__(self, filename: str):
    """Initializes the LogFile.

    Args:
      filename: the leveldb log filename
    """
    self.filename = filename

  def GetBlocks(self) -> Generator[Block, None, None]:
    """Returns an iterator of Block instances.

    A logfile is composed of one or more blocks.

    Yields:
      a Block
    """
    with open(self.filename, 'rb') as fh:
      block = Block.FromStream(fh)
      while block:
        yield block
        block = Block.FromStream(fh)

  def GetPhysicalRecords(self) -> Generator[PhysicalRecord, None, None]:
    """Returns an iterator of PhysicalRecord instances.

    A block is composed of one or more physical records.

    Yields:
      PhysicalRecord
    """
    for block in self.GetBlocks():
      yield from block.GetPhysicalRecords()

  def GetWriteBatches(self) -> Generator[WriteBatch, None, None]:
    """Returns an iterator of WriteBatch instances.

    Depending on the batch size, a log file batch can be spread across one or
    more physical records.

    Yields:
      WriteBatch
    """
    buffer = bytearray()
    for physical_record in self.GetPhysicalRecords():
      if (physical_record.record_type ==
         definitions.LogFilePhysicalRecordType.FULL):
        buffer = physical_record.contents
        offset = physical_record.contents_offset + physical_record.base_offset
        yield WriteBatch.FromBytes(buffer, base_offset=offset)
        buffer = bytearray()
      elif (physical_record.record_type
            == definitions.LogFilePhysicalRecordType.FIRST):
        offset = physical_record.contents_offset + physical_record.base_offset
        buffer = bytearray(physical_record.contents)
      elif (physical_record.record_type ==
            definitions.LogFilePhysicalRecordType.MIDDLE):
        buffer.extend(bytearray(physical_record.contents))
      elif (physical_record.record_type ==
            definitions.LogFilePhysicalRecordType.LAST):
        buffer.extend(bytearray(physical_record.contents))
        yield WriteBatch.FromBytes(buffer, base_offset=offset)
        buffer = bytearray()

  def GetParsedInternalKeys(self) -> Generator[ParsedInternalKey, None, None]:
    """Returns an iterator of ParsedInternalKey instances.

    A batch can contain one or more key value records.

    Yields:
      ParsedInternalKey
    """
    for batch in self.GetWriteBatches():
      yield from batch.records
